// Copyright - Simovits Consulting 2009
// Author: James Dickson.

#include "common.h"
#include <openssl/err.h>
#include <string.h>

#include <openssl/err.h>
#include <openssl/ssl.h>
#include <openssl/x509v3.h>


BIO *bio_err=0;
SSL_METHOD *getMethod(RequestProtocol protocol);


/**
 * return the grade of the connection method
 */
int getGradeByProtocolAndBits(RequestProtocol protocol, 
							  int bits, 
							  char *strCipherDescription, 
							  BOOL nessusCipherTest,
							  char *strCipherName)
{
	// James Note: MD5 has some weaknesses. However when used for 
	// SSL communication design it should not be a PCI compliancy factor.
	// However future ciphers may affect Nessus-implementation correctness
	// thus, as a reminder I added the nessusCipherTest-flag.

	/*
		[PCI-DSS Glossary definition of Strong Cryptography, 2008]

		"Cryptography based on industry-tested and accepted algorithms, along with
		strong key lengths and proper key-management practices. Cryptography is a
		method to protect data and includes both encryption (which is reversible) and
		hashing (which is not reversible, or one way). 
		
		SHA-1 is an example of an industry-tested and accepted hashing algorithm. 
		
		Examples of industry-tested and accepted standards and algorithms for encryption include 
		 - AES (128 bits and higher), 
		 - TDES (minimum double-length keys (james note: 112 effective bits)), 
		 - RSA (1024 bits and higher),
		 - ECC (160 bits and higher), and 
		 - ElGamal (1024 bits and higher).

		See NIST Special Publication 800-57 (http://csrc.nist.gov/publications/) for
		more information."
	*/

	#define STRLOW "WEAK"
	#define STRMEDIUM "INTERMEDIATE"
	#define STRHIGH "STRONG"

	// SSL2 is always LOW
	if(protocol == SSL2 || bits < 56)
	{
		strcpy(strCipherDescription, STRLOW);
		return 1;
	}	

	// Export, Anonymous or NULL ciphers are always LOW
	char *strADH = strstr(strCipherName, "ADH");
	char *strNULL = strstr(strCipherName, "NULL");
	char *strEXP = strstr(strCipherName, "EXP");

	if(strADH || strNULL || strEXP)
	{
		strcpy(strCipherDescription, STRLOW);
		return 1;
	}
	
	if(nessusCipherTest)
	{
		if(bits < 112 ) 
		{
			strcpy(strCipherDescription, STRMEDIUM);
			return 2;
		}
	}
	else
	{
		char *strAES = strstr(strCipherName, "AES");
		char *strDES = strstr(strCipherName, "DES");

		if(strAES != NULL)
		{		
			if(bits < 128 ) 
			{
				strcpy(strCipherDescription, STRMEDIUM);
				return 2;
			}
		}
		else if(strDES != NULL)
		{
			if(bits < 112 ) 
			{
				strcpy(strCipherDescription, STRMEDIUM);
				return 2;
			}
		}
		else 
		{
			if(bits < 128 ) 
			{
				strcpy(strCipherDescription, STRMEDIUM);
				return 2;
			}
		}
	}

	strcpy(strCipherDescription, STRHIGH);
	return 3;
}

/**
 * Convert string to upper case
 */
void strtoupper(char *str, int maxlen)
{
	int i = 0 ;

	while(i < maxlen && str[i] != 0)
	{
		str[i] = toupper(str[i]); 
		i++ ;
	}
}

BOOL getCertAlgortithm(X509 *crtCurrent, char *buf, int buflen)
{
	
	// Signature algorithm
	if (!(X509_FLAG_COMPAT & X509_FLAG_NO_SIGNAME))
	{
		int ret = i2t_ASN1_OBJECT(buf, buflen, crtCurrent->cert_info->signature->algorithm);		

		if(ret > 0)
		{
			return TRUE;
		}
	}

	return FALSE;
}

/**
 * Print the cert to stdout.
 */
void printCert(X509 *crtCurrent)
{
	#define MAXBUF 2048
	BIO *stdConsole = BIO_new(BIO_s_file());
	BIO_set_fp(stdConsole, stdout, BIO_NOCLOSE);
	char subjectBuffer[MAXBUF];
	char buffer[MAXBUF];

	if(crtCurrent == NULL) return;

	printf("\n\n");
	printf("[+] Certificate:\n-----------------\n\n");

	// Subject
	if (!(X509_FLAG_COMPAT & X509_FLAG_NO_SUBJECT))
	{
		X509_NAME_oneline(X509_get_subject_name(crtCurrent), subjectBuffer, MAXBUF);
		printf("\t* Subject: %s\n", subjectBuffer);
	}
	
	// Issuer
	if (!(X509_FLAG_COMPAT & X509_FLAG_NO_ISSUER))
	{		
		X509_NAME_oneline(X509_get_issuer_name(crtCurrent), buffer, MAXBUF);
		printf("\t* Issuer: %s\n", buffer);		

	}

	// Validity
	if (!(X509_FLAG_COMPAT & X509_FLAG_NO_VALIDITY))
	{
		printf("\t* Not valid before: ");
		ASN1_TIME_print(stdConsole, X509_get_notBefore(crtCurrent));

		printf("\n");
		printf("\t* Not valid after: ");
		ASN1_TIME_print(stdConsole, X509_get_notAfter(crtCurrent));
		printf("\n");
	}

	// Cert Version
	if (!(X509_FLAG_COMPAT & X509_FLAG_NO_VERSION))
	{
		long version = X509_get_version(crtCurrent);
		printf("\t* Version: %lu\n", version);
	}

	// Cert Serial No.
	if (!(X509_FLAG_COMPAT & X509_FLAG_NO_SERIAL))
	{
		long serial = ASN1_INTEGER_get(X509_get_serialNumber(crtCurrent));
		printf("\t* Serial number: %lu  (%02X)\n", serial, serial);
	}

	// Signature algorithm
	if (!(X509_FLAG_COMPAT & X509_FLAG_NO_SIGNAME))
	{
		printf("\t* Signature algorithm: ");
		i2a_ASN1_OBJECT(stdConsole, crtCurrent->cert_info->signature->algorithm);
		printf("\n");
	}

	// Public key algorithm
	if (!(X509_FLAG_COMPAT & X509_FLAG_NO_PUBKEY))
	{
		printf("\t* Public Key Algorithm: ");
		i2a_ASN1_OBJECT(stdConsole, crtCurrent->cert_info->key->algor->algorithm);
		printf("\n");

		
		// Public key
		EVP_PKEY *pubKey = X509_get_pubkey(crtCurrent);

		if (pubKey != NULL)
		{					
			printf("\t* Public key type: ");

			switch (pubKey->type)
			{
				case EVP_PKEY_RSA:
					printf("RSA Public key: (%d bit)\n", BN_num_bits(pubKey->pkey.rsa->n));
					RSA_print(stdConsole, pubKey->pkey.rsa, 6);

				break;
			
				case EVP_PKEY_DSA:
					printf("DSA Public key:\n");				
					DSA_print(stdConsole, pubKey->pkey.dsa, 6);
				break;

				case EVP_PKEY_EC:
					printf("EC Public key:\n");
					EC_KEY_print(stdConsole, pubKey->pkey.ec, 6);
				break;

				default:
					printf("\t\t* Unknown public key type: %d\n", pubKey->type);
				break;
			}

			EVP_PKEY_free(pubKey);
		}
		else
		{
			printf("[-] ERROR: X509_get_pubkey() failed.\n");
		}
	}

	//crtCurrent->ex_kusage

	// X509v3
	if (!(X509_FLAG_COMPAT & X509_FLAG_NO_EXTENSIONS))
	{
		int extensionCount = sk_X509_EXTENSION_num(crtCurrent->cert_info->extensions);

		if (extensionCount > 0)
		{
			printf("\t* X509v3 extensions:\n");

			for (int i = 0; i < extensionCount; i++)
			{
				X509_EXTENSION *extension = sk_X509_EXTENSION_value(crtCurrent->cert_info->extensions, i);
				
				//BASIC_CONSTRAINTS *bs = (BASIC_CONSTRAINTS *) X509V3_EXT_d2i(extension);


				printf("\t* ");

				/*
				if(bs != NULL)
				{
					printf("Basic constraint: CA=%s, Pathlength=%d - ", (bs->ca ? "True" : "False"), bs->pathlen);
				}
				*/

				// Extension name
				ASN1_OBJECT* aobj = X509_EXTENSION_get_object(extension);
				i2a_ASN1_OBJECT(stdConsole, aobj);

				// Check if critical
				int critical = X509_EXTENSION_get_critical(extension);
				if(critical)
				{
					printf(" [Critical (%d)] ", critical);
				}
	
				// Value
				printf(" - value:\n");
				if (!X509V3_EXT_print(stdConsole, extension, X509_FLAG_COMPAT, 8))
				{
					M_ASN1_OCTET_STRING_print(stdConsole, extension->value);
				}		

				printf("\n");
			}

		}
	}

	printf("-----------------\n\n");
}

/**
 * Get a list of all valid ciphers.
 */
int getAvailableCiphers(CIPHERTESTITEM *cipherItems, int maxItems, TESTINFO testInfo)
{
	const int METHOD_COUNT = 4;
	int current = 0;
	SSL_METHOD *vMethods[METHOD_COUNT];
	RequestProtocol protocols[METHOD_COUNT];
	
	// Init the methods.
	vMethods[0] = (SSL_METHOD *) SSLv2_client_method();
	protocols[0] = SSL2;

	vMethods[1] = (SSL_METHOD *) SSLv3_client_method();
	protocols[1] = SSL3;

	vMethods[2] = (SSL_METHOD *) TLSv1_client_method();
	protocols[2] = TLS1;

	vMethods[3] = (SSL_METHOD *) DTLSv1_client_method();
	protocols[3] = DTLS1;				

	// Go thru all possible methods and place them in the list.
	for(int x=0; x < METHOD_COUNT; x++)
	{
		SSL_CTX *ctx = SSL_CTX_new(vMethods[x]);

		if (ctx != NULL)
		{
			SSL *ssl = SSL_new(ctx);
			SSL_CTX_set_cipher_list(ctx, "ALL:COMPLEMENTOFALL");					

			if (ssl != NULL)
			{
				STACK_OF(SSL_CIPHER) *clist = SSL_get_ciphers(ssl);
		
				if(clist != NULL)
				{
					int cipherCount = sk_SSL_CIPHER_num(clist);

					for (int i = 0; i < cipherCount; i++)
					{

						if(current < maxItems)
						{
							cipherItems[current].protocol = protocols[x];

							char * str = (char *) SSL_CIPHER_get_name(sk_SSL_CIPHER_value(clist, i));

							strcpy(cipherItems[current].cinf.strOpenSSLShortName, str);

							SSL_CIPHER_description(sk_SSL_CIPHER_value(clist, i), cipherItems[current].cinf.strOpenSSLName, 255);		

							int len = strlen(cipherItems[current].cinf.strOpenSSLName);
							cipherItems[current].cinf.strOpenSSLName[len-2] = 0;

							cipherItems[current].cinf.bits = SSL_CIPHER_get_bits(sk_SSL_CIPHER_value(clist, i), &cipherItems[current].cinf.bits);
							cipherItems[current].strVersion = SSL_CIPHER_get_version(sk_SSL_CIPHER_value(clist, i));	
							cipherItems[current].testCount = 0;

							

							// Get the strength							
							int strength = getGradeByProtocolAndBits(
								protocols[x], 
								cipherItems[current].cinf.bits, 
								cipherItems[current].cinf.strStrengthName, 
								testInfo.bUseNessusCipherTest,
								cipherItems[current].cinf.strOpenSSLShortName);

							cipherItems[current].cinf.strength = strength;
							

							current++;
						}
						else
						{
							printf("[-] ERROR: Could not get info about ciphers. To little space allocated \n");
							exit(-1);
						}
					}
				}
				else
				{
					printf("[-] ERROR: Could not init cipher list\n");
				}
		
				
				SSL_free(ssl); // Cleanup
			}
			else
			{
				current = -1;
				printf("[-] ERROR: Could not init SSL\n");
			}
			
			SSL_CTX_free(ctx);	// Cleanup
		}
		else
		{
			current = -2;
			printf("[-] ERROR: Could not init CTX.");
		}
	}

	return current;
}



// Init test info
void initTestInfo(TESTINFO *tInf)
{
	tInf->debugLevel =		0;
	tInf->strHostname =		NULL;
	tInf->port =			443;
	tInf->strIPAdress =		NULL;
	tInf->bTestSend =		FALSE;
	tInf->strUseragent =	"Mozilla";
	tInf->strGetHostname =	NULL;
	tInf->bVerbose =		FALSE;
	tInf->bSuppressSummary = FALSE;
	tInf->bCheckCert =		TRUE;
	tInf->strCACertPath =	NULL;
	tInf->strServerCertPath = NULL;
	tInf->strRequest =		NULL;
	tInf->secondTimeout =	5000;
	tInf->strSIPDomain =	NULL;
	tInf->testSIP =			FALSE;
	tInf->bPrintCommunication = FALSE;
	tInf->bTestOnlySpecificCryptos = FALSE;
	tInf->bTestOnlySpecificProtocols = FALSE;
	tInf->bServerMode =		FALSE;
	tInf->serverProtocol =	TLS1;
	tInf->strServerCertPassword = "password";
	tInf->strServerCipher = NULL;
	tInf->strServerCertKeyfile = NULL;
	tInf->strHTTPPage =		"/";
	tInf->msInterval =		0;
	tInf->onlyGrade =		-1;
	tInf->bUDP =			FALSE;
	tInf->bExitOnBitMismatch = FALSE;
	tInf->bUseOpenSSLBits = TRUE;
	tInf->bPrintAvailableCiphers = FALSE;
	tInf->bOnlyAvailableCiphers = TRUE;
	tInf->bPrintCert = FALSE;
	tInf->bSuppressUserSupplied = FALSE;
	tInf->bStartTLS = FALSE;
	tInf->bUseWorkarounds = TRUE;
	tInf->bUseNessusCipherTest = TRUE;
	tInf->bTestRenegotiation = TRUE;
	tInf->bUseColors = TRUE;
	tInf->bUseCiscoSpeshul = FALSE;

	strcpy(tInf->strSMTPConnect, "SSLDiagnos.simovits.com");
	
}

/**
 * Print usage...
 */
void printUsage()
{
	printf(

"\n................\n\n"

"--testsend\n"
"If specified then when connected to the server an init-message is sent. \n"
"If --testsip is also specified then a SIP-message is sent otherwise an HTTP-GET"
"is sent. \n\n"


"--testsip \n"
"If specified then the init-message uses a SIP-message.\n\n"

"--readtimeout <timeout> \n"
"Read timeout in milliseconds (default=5000) \n\n"

"--certout <path>\n"
"If specified then the certificate is stored at the path. It is overwritten for every time a connection is successful \n\n"

"--page <http-page> \n"
"Allows the user specify what the http-get page should be used (if the url is https://test.se/foo.htm then specify --page foo.htm \n\n"

"-pr , --printcom \n" 
"If specified the communication is printed to standard out\n\n"

"--servermode\n" 
"Server mode should be used. EXPERIMENTAL\n\n"

"-t <targetname>\n"
"The server to connect to\n\n"

"--printavailable\n"
"Print available ciphers\n\n"

"--interval <ms of interval>\n"
"The number of ms to sleep between every test (default=0)\n\n"

"-h <hostname>	\n"
"When --testsend is specifed and a HTTP-GET is sent to the server then this specifes"
"what hostname-header that should be sent.\n\n"

"--servercipher <cipherlist> \n"
"Specifies what ciphers (OpenSSL-style) that should be used for the server mode\n\n"

	
"--serverprotocol <protocol>\n"
"Specifies what protocols that are used for server mode (SSL2, SSL3 or TLS1).\n\n"

"--protocol <protocol>\n"
"Specifies what protocols that are used for client mode (SSL2, SSL3 or TLS1). \n\n"

"--cipher <cipherlist> \n"
"Specifies what ciphers (OpenSSL-style) that should be used for the client mode. "
"If specified then only specific ciphers are tested. It can be specified more than once\n\n"

"-c <path> \n"
"Specifies the path to the certificate (client mode) \n\n"

"--servercert <path>\n"
"Specifies the path to the certificate (server mode) \n\n"

"--servercertkey <path> \n"
"Specifies the path to the certificate key file (server mode) \n\n"

"--servercertpass <password>\n"
"Specifies the password for the certificate key file (server mode)\n\n"

"--sipdomain <domain>\n"
"Specifies what domain is to be used with SIP-requests \n\n"

"-r <request>\n"
"Used to specify a custom request-message (instead of HTTP-GET or SIP). \n\n"

"-p <port>\n"
"The port number to use \n\n"

"--help\n"
"This help \n\n"

"--printcert\n"
"Print certificate information\n\n"

"--smtptls\n"
"Use this if you need to test the TLS on an SMTP-server. This sets up the TLS using STARTTLS first. \n\n"

"--disableworkarounds\n"
"Disable SSL-workarounds\n\n"

"--nocolor\n"
"Disable color usage.\n\n"

"--ciscospeshul\n"
"Use Cisco AnyConnect speshul dtls\n\n"

"--strictpcigrade\n"
"Use strict PCI grade for ciphers instead of the way Nessus grades ciphers Low, Medium, High. Some potential difference for AES. \n\n"

"--popssl\n"
"Starts POP3 SSL (using command STLS). example: ssldiagnos -t target --popssl -p 110\n\n"

"--UDP\n"
"If DTLS should be tested. Does not seem to work with Cisco AnyConnect (handshake failure)\n\n");



}



/**
 * Get cert. type string
 */
int getCertTypeString(int type, char *buf)
{
	switch (type)
	{
		case EVP_PKEY_RSA:
			strcpy(buf, "RSA");
			break;

		case EVP_PKEY_DSA:
			strcpy(buf, "DSA");
			break;

		case EVP_PKEY_EC:
			strcpy(buf, "Elliptic Curve");
			break;
		default:

			strcpy(buf, "Unknown");
			 
		break;
	}

	return 1;
}


/**
 * Get cert. bitcount.
 */
int getCertBitCount(X509 *x509Cert)
{
	EVP_PKEY *publicKey = X509_get_pubkey(x509Cert);

	int bits = -1;

	if (publicKey == NULL)
	{
		return -1;
	}
	else
	{
		
		switch (publicKey->type)
		{
			case EVP_PKEY_RSA:
			{
				bits = BN_num_bits(publicKey->pkey.rsa->n);		
			}
			break;

			case EVP_PKEY_DSA:
			{
				bits = BN_num_bits(publicKey->pkey.dsa->p);
			}
			break;

			case EVP_PKEY_EC:
			{
				BIGNUM *order = BN_new();

				if (!order)
				{
					ERR_clear_error();
					printf("[-] ERROR: Could not get new bignum while trying to figure out elliptic curve bitcount!\n");
					return -1;
				}

				const EC_GROUP *grp = (const EC_GROUP*) EC_KEY_get0_group(publicKey->pkey.ec);

				if (!EC_GROUP_get_order(grp, order, NULL))
				{
					ERR_clear_error();
					return 0;
				}

				bits = BN_num_bits(order);
				BN_free(order);

			}
			break;

			default:
			break;
		}

		EVP_PKEY_free(publicKey);
	}	

	return bits;
}



     
/**
 * @return 0 if not alpha numeric otherwise 1.
 */
int isInt(char *str, int len)
{
	for(int i=0; i < len; i++)
	{
		if(! (str[i] >= '0' && str[i] <= '9'))
		{
			return 0;
		}
	}

	return 1;
}

/**
 * Init CTX
 */
SSL_METHOD *getMethod(RequestProtocol protocol)
{
    SSL_METHOD *meth;
    SSL_CTX *ctx;

	if(protocol == SSL2)
	{
		meth = (SSL_METHOD *) SSLv2_method();
	}
	else if(protocol == SSL3)
	{
		meth = (SSL_METHOD *) SSLv3_method();
	}
	else if(protocol == SSL2OR3)
	{
		meth = (SSL_METHOD *)SSLv23_method();
	}
	else if(protocol == TLS1)
	{
		meth = (SSL_METHOD *) TLSv1_method();
	}
	else if(protocol == DTLS1)	// 2010-03-24 - james - We also want to be able to test Datagram TLS (UDP).
	{
		meth = (SSL_METHOD *) DTLSv1_client_method();
	}

	return meth;
    
}

/**
 * Init CTX
 */
SSL_CTX *initCTX(RequestProtocol protocol, BOOL bUseWorkarounds, BOOL bUseCiscoSpeshul)
{
    SSL_METHOD *meth = getMethod(protocol);
    SSL_CTX *ctx;
    
    ctx=SSL_CTX_new(meth);

	DWORD currentOptions = SSL_CTX_get_options(ctx); 

	if(bUseWorkarounds)
	{
		// Use all workarounds.				
		// 2010-11-23 - james - Add this since we do not want to remove anything and always allow unsafe renegotiation.
		// In practice this code has little meaning.

		currentOptions = currentOptions | SSL_OP_ALL; 	
		currentOptions = currentOptions | SSL_OP_LEGACY_SERVER_CONNECT;
		currentOptions = currentOptions | SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION;
	}

	if(bUseCiscoSpeshul)
	{
		currentOptions = currentOptions | SSL_OP_CISCO_ANYCONNECT; 		
	}

	// 2010-03-24 - james - Required for DTLS - please refer apps/s_client.c source file
	if(protocol == DTLS1)
	{
		SSL_CTX_set_read_ahead (ctx, 1);
	}

	// Set options
	SSL_CTX_set_options(ctx, currentOptions);

    
    return ctx;
}
